-- Pylon: a foundation library; version 0.81(beta); 1997-05-08

-- 
-- Copyright (c) 1997 AlterSoft, England, E.U.
-- 
-- Permission is hereby granted, without written agreement and without
-- licence or royalty fees, to use, copy, modify, or distribute this source
-- code or binary programs using it or part of it, subject to the following
-- condition:
-- 
-- The above copyright notice and the following disclaimer must appear in
-- all copies or substantial portions of this source code. (Binary programs
-- derived from the source code can be distributed without restriction or
-- mention of the origin of the source.)
-- 
-- THIS SOFTWARE IS PROVIDED BY ALTERSOFT AND THE AUTHORS ``AS IS'' AND
-- ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED. IN NO EVENT SHALL ALTERSOFT OR THE AUTHORS BE LIABLE
-- TO ANY PARTY FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES ARISING IN ANY WAY OUT OF THE USE OF THIS SOURCE
-- CODE.
-- 

indexing
	cluster: Pylon, Format
	description: "Decode string to integers in base 2-36"
	interface: client

class P_INTEGER_DECODER

creation 
	make
	
feature -- Creation
	
	make(new_base : INTEGER) is
		require
			ok: new_base > 2
		do
			check max_base: base <= digit_table.count end
			base := new_base
		end

feature -- Base

	base : INTEGER

feature -- Integer

	extend(digit : CHARACTER) is
			-- Add right digit.
		require
			base: is_digit(digit)
		do
			value := (value * base) + char_to_int(digit)
			digits := digits + 1
		ensure
			done_integer: value = ((old value) * base + char_to_int(digit))
			done_digits: digits = old digits + 1
		end

	reset is
		do
			value := 0
			digits := 0
		ensure
			ok: value = 0 and digits = 0
		end

	value : INTEGER
			-- Current integer.

	digits : INTEGER
			-- Number of converted digits (not including sign).

feature -- String to integer

	from_string(str : STRING) is
		require
			ok: str /= Void
		local
			i : INTEGER
			minus : BOOLEAN
		do
			reset
			if str.count > 0 then
				i := 1
				
				-- check sign
				if str.item(1) = '+' then
					i := 2
				elseif str.item(1) = '-' then
					minus := true
					i := 2
				end
				
				-- digits
				from 
				variant 
					str.count + 1 - i 
				until
					i > str.count or else not is_digit(str.item(i)) 
				loop
					extend(str.item(i))
					i := i + 1
				end
				
				-- restore sign
				if minus then
					value := - value
				end
			end
		end
	
feature -- Digit level

	int_to_char(digit : INTEGER) : CHARACTER is
			-- Integer to character digit.
		require
			ok: digit >= 0 and digit < base
		do
			Result := digit_table.item(digit+1)
		ensure
			good_base: is_digit(Result)
			done: char_to_int(Result) = digit 
		end

	char_to_int(c : CHARACTER) : INTEGER is
			-- Character to integer.
		require
			is_intchar: is_digit(c)
		do
			Result := internal_char_to_int(c)
		ensure
			good_base: Result >= 0 and Result < base
			done: -- (case sensitive) int_to_char(Result) = c
		end

	is_digit(c : CHARACTER) : BOOLEAN is
			-- Is this character a valid digit?
		local
			int : INTEGER
		do
			int := internal_char_to_int(c)
			Result := (int >= 0) and (int < base)
		end

feature -- Encode

	value_to_string(int : INTEGER) : STRING is
			-- Encode character.
		local
			i : INTEGER
			fmt : P_FORMAT
		do
			!!fmt
			!!Result.make(0)
			
			if int = 0 then
				Result.append_character('0')
			else
				from
					if int < 0 then i := - int else i := int end
					check positive: i >= 0 end
				until i = 0 loop
					Result.append_character(int_to_char(i \\ base))
					i := i // base
				end
			end

			if int < 0 then
				Result.append_character('-')
			end
		
			fmt.invert(Result)
		ensure
			is_copy: Result /= Void
			done: Result.count > 0
		end

feature { NONE }
	
	internal_char_to_int(c : CHARACTER) : INTEGER is
		local
			i : INTEGER
		do
			-- Slow but charset independent method.
			Result := -1
			
			from i := 1 variant digit_table.count + 1 - i until 
				Result > 0 or i > digit_table.count
			loop
				if c = digit_table.item(i) or else c = digit_upper_table.item(i) then
					Result := i - 1
				end
				i := i + 1
			end
		end

	digit_table : STRING is "0123456789abcdefghijklmnopqrstuvwxyz"
			-- item(1) = 0, item(2) = 1, etc, could be customised

	digit_upper_table : STRING is "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
			--? No standard CHARACTER.[to_]lower...

invariant
	base_ok: base >= 2 and base <= digit_table.count

end
