-- Pylon: a foundation library; version 0.81(beta); 1997-05-08

-- 
-- Copyright (c) 1997 AlterSoft, England, E.U.
-- 
-- Permission is hereby granted, without written agreement and without
-- licence or royalty fees, to use, copy, modify, or distribute this source
-- code or binary programs using it or part of it, subject to the following
-- condition:
-- 
-- The above copyright notice and the following disclaimer must appear in
-- all copies or substantial portions of this source code. (Binary programs
-- derived from the source code can be distributed without restriction or
-- mention of the origin of the source.)
-- 
-- THIS SOFTWARE IS PROVIDED BY ALTERSOFT AND THE AUTHORS ``AS IS'' AND
-- ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED. IN NO EVENT SHALL ALTERSOFT OR THE AUTHORS BE LIABLE
-- TO ANY PARTY FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES ARISING IN ANY WAY OUT OF THE USE OF THIS SOURCE
-- CODE.
-- 

indexing
	cluster: Pylon, Container
	description: "Abstract list (traversable collection)"
	interface: abstract

deferred class P_LIST[G]

inherit
	P_TRAVERSABLE[G]
		end
	P_SORTABLE[G]
		end
	P_SEARCHABLE[G,G]
		rename
			search as value_search
		end

feature { NONE } -- Attributes

	found_result : BOOLEAN
			-- Item has been found.
	found_quantity : INTEGER
			-- Quantity found.

feature -- Update 

	add(it : G) is
			-- Add item to the container.
		require
			writable: is_writable
			not_void: not is_void(it)
		deferred
		ensure
			keep_reference: item = it
		end
	
	replace(it : G) is
			-- Replace the current item.
		require
			writable: is_writable
			not_empty: not is_empty
			not_void: not is_void(it)
		deferred
		ensure
			keep_reference: item = it
		end

	insert(it : G) is
			-- Insert item before the current item.
		require
			writable: is_writable
			not_empty: not is_empty
			not_void: not is_void(it)
		deferred
		ensure
			keep_reference: item = it
		end

	remove is
			-- Remove the current item.
		require
			writable: is_writable
			not_empty: not is_empty
			unlocked: not is_locked
		deferred
		end

feature -- Duplication

	copy(other: like Current) is
			-- GENERAL's copy.
			-- Time: O(other.count)
		do
			reset
			append(other)
		end

feature -- Comparison

	is_equal(other : like Current) : BOOLEAN is
			-- GENERAL's equality relation (requires covariant arguments).
			-- Time: O(count)
		do
			Result := is_equal_list(other)
		end

	is_equal_list(other : P_LIST[G]) : BOOLEAN is
			-- Compare two lists which can be of different dynamic type.
			-- Time: O(count)
		require
			valid: other /= Void
		local
			it,it_other : P_ITERATOR[G]
		do
			if other.count = count then
				Result := true
				
				from
					it := iterator
					it.first
					it_other := other.iterator
					it_other.first
				invariant
					synch: it.forward_variant = it_other.forward_variant
				variant
					it.forward_variant
				until
					it.outside or Result = false -- it_other.outside checked with count 
				loop
					Result := (it.item).is_equal(it_other.item)
					
					-- next
					it.forth
					it_other.forth
				end
			end
		end

feature -- Conversion

	from_list(other : P_LIST[G]) is
			-- Convert from another list.
			-- Note: The previous contents of this list is deleted.
			-- Time: O(other.count)
		require
			valid: other /= Void
		do
			reset
			append(other)
		ensure
			done: is_equal_list(other)
			keep_reference: -- keep copy of items.
		end

	from_array(other : ARRAY[G]) is
			-- Convert from an ARRAY.
			-- Note: The previous contents of this list is deleted.
			-- Time: O(other.count)
		require
			valid: other /= Void
			not_void: -- for_all other.item, other.item /= Void
		local
			i : INTEGER
		do
			reset
			from
				i := other.lower
			invariant
				bounds: other.lower <= other.upper implies (i >= other.lower and i <= other.upper + 1)
			variant
				other.upper - i + 1
			until
				i > other.upper
			loop
				add(other.item(i))
				i := i + 1
			end
		ensure
			done: other.count = count
			keep_reference: -- keep copy of items.
		end

	from_iterable(it : P_ONEWAY_ITERATOR[G]) is
			-- Convert from a one way linear iterable container.
			-- Note: The previous contents of this list is deleted.
			-- Time: O("it.count")
		require
			valid: it /= Void
			initialised: it.outside
		do
			reset
			append_iterable(it)
		ensure
			keep_reference: -- keep copy of items.
		end

	to_array : ARRAY[G] is
			-- Convert list to ARRAY.
			-- Time: O(count)
		local
			it : P_ITERATOR[G];
			i : INTEGER
		do
			if count > 0 then 
				!!Result.make(1,count)
				from
					it := iterator
					it.first
					i := 1
				variant
					it.forward_variant
				until
					it.outside
				loop
					Result.put(it.item,i)
					i := i + 1
					it.forth
				end
			end
		ensure
			done: is_empty = (Result = Void)
			is_shallow_copy: -- new array, old items.
		end

feature -- Concatenation

	append(other : P_LIST[G]) is
			-- Append other container to current.
			-- Time: O(other.count)
		require
			valid: other /= Void
		do
			append_iterable(other.iterator)
		ensure
			keep_reference: -- keep copy of items.
		end

	append_iterable(it : P_ONEWAY_ITERATOR[G]) is
			-- Append the items from a linear iterator.
			-- Time: O("it.count")
		require
			valid_iterator: it /= Void
			initialised: it.outside
		do
			from
				it.first
			variant
				it.forward_variant
			until
				it.outside
			loop
				add(it.item)
				
				-- next
				it.forth
			end
		ensure
			keep_reference: -- keep copy of items.
		end

feature -- Search
	
	equality_search(x: G) is
			-- Search items in the container which are equal (as defined 
			-- by the = operator) with 'x'.
			-- Time: O(count)
		require
			readable: is_readable
		local
			it : P_ITERATOR[G]
		do
			found_result := false
			if count > 0 then
				from
					it := iterator
					it.first
					found_quantity := 0
				variant
					it.forward_variant
				until
					it.outside
				loop
					if it.item = x then
						found_result := true
						if found_quantity = 0 then
							it.synchronize_container
						end
						found_quantity := found_quantity + 1
					end
					it.forth
				end
			end
		end

	value_search(x : G) is
			-- Search items in the container which are of equal value (as defined 
			-- by the is_equal routine) with 'x'.
			-- Time: O(count)
		local
			it : P_ITERATOR[G]
		do
			found_result := false
			if count > 0 then
				from
					it := iterator
					it.first
					found_quantity := 0
				variant
					it.forward_variant
				until
					it.outside
				loop
					if (it.item).is_equal(x) then
						found_result := true
						if found_quantity = 0 then
							it.synchronize_container
						end
						found_quantity := found_quantity + 1
					end
					it.forth
				end
			end
		end

	found : BOOLEAN is
		do
			Result := found_result
		end

	found_count : INTEGER is
		do
			Result := found_quantity
		end

feature -- Sorting

	is_sorted : BOOLEAN is
			-- Is the container sorted?
			-- Time: O(target.count)
		local
			it : P_ITERATOR[G];
			previous_item : G
		do
			from
				Result := true
				it := iterator
				it.first
				if not it.outside then
					previous_item := it.item
					it.forth
				end
			invariant
				previous: -- not it.outside implies (previous_item valid)
			variant
				it.forward_variant
			until
				it.outside or else Result = false
			loop
				Result := order_relation.ordered(previous_item,it.item)
				it.forth
			end
			-- Note: empty container or singleton are always sorted.
		end
	
end
