-- Pylon: a foundation library; version 0.81(beta); 1997-05-08

-- 
-- Copyright (c) 1997 AlterSoft, England, E.U.
-- 
-- Permission is hereby granted, without written agreement and without
-- licence or royalty fees, to use, copy, modify, or distribute this source
-- code or binary programs using it or part of it, subject to the following
-- condition:
-- 
-- The above copyright notice and the following disclaimer must appear in
-- all copies or substantial portions of this source code. (Binary programs
-- derived from the source code can be distributed without restriction or
-- mention of the origin of the source.)
-- 
-- THIS SOFTWARE IS PROVIDED BY ALTERSOFT AND THE AUTHORS ``AS IS'' AND
-- ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED. IN NO EVENT SHALL ALTERSOFT OR THE AUTHORS BE LIABLE
-- TO ANY PARTY FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES ARISING IN ANY WAY OUT OF THE USE OF THIS SOURCE
-- CODE.
-- 

indexing
	cluster: Pylon
	description: "Math functions missing from kernel and used by this library"
	comment: "Implementation is typically very inefficient (use only when performance not an issue)"
	interface: mixin, client, internal

class P_MATH

feature -- Math

	e : REAL is 2.71828182845904523536 
	
	int_sqrt(n : INTEGER) : INTEGER is
			-- Calculate integer square root.
			-- Time: O(sqrt(N)) 
		require
			postive: n >= 0
		local
			u, v : INTEGER 
		do
			if n < 2 then
				Result := n
			else
				------------------------------------
				-- u(n) = n^2
				-- v(n) = (n+1)^2 - n^2 = 2n + 1
				------------------------------------
				-- u(n+1) = u(n) + v(n)
				-- v(n+1) = v(n) + 2
				------------------------------------
				from
					Result := 1 -- Result = (n-1)
					u := 4
					v := 5
				until
					u > n
				loop
					u := u + v
					v := v + 2
					Result := Result + 1
				end
			end
		ensure
			ok: ((Result^2) <= n) and ((Result+1)^2 > n)
		end

	square_root(x : DOUBLE) : DOUBLE is
			-- Square root (floating point).
			-- Time: O(1) for x in 0.001..100, O(n) for x = a.10^n, slow in any case.
		require
			positive: x >= 0
		local
			xp,res : DOUBLE
			exp,n : INTEGER
		do
			if x > 0 then -- else x = 0 implies Result = 0
				xp := x

				-- Reduce to (0.001,100) interval, so that:
				-- x = xp * (10 ^ (2*n))
				from until xp <= 100 loop
					exp := exp + 1
					xp := xp / 100.0
				end
				from until xp >= 0.001 loop
				    exp := exp - 1
				    xp := xp * 100.0
				end
				check done_interval: xp <= 100 and xp >= 0.001 end

				-- Now get square root of xp thanks to f(x) = 1/2 * ( x + C/x )
				-- which converges to square_root(C).
				from 
					res := 1
				until n > 10 loop
					res := ((res + xp / res) / 2)
					n := n + 1
				end

				-- Final result
				Result := res * (10 ^ exp)
				check precision: ((res*res) - (x/(10^exp))) < 0.001 end
			end
		ensure
			positive: Result >= 0
			done: -- (Result*Result ~= x)
		end
	
feature -- Trigonometry
	
	Pi_value : DOUBLE is 3.1415926535897932384626

	sine(x : DOUBLE) : DOUBLE is
			-- Portable sine.
			-- Time: O(1) but far too slow.
		local
			bx,pi2 : DOUBLE
			n : INTEGER
			term_sign : INTEGER
		do
			pi2 := Pi_value * 2.0
			bx := mod(x.abs,pi2) * x.sign
			check bounds: bx >= - pi2 and bx <= pi2 end
			
			if bx > Pi_value then -- get in [-Pi,Pi]
				bx := bx - pi2
			elseif bx < -Pi_value then
				bx := bx + pi2
			end
			if bx > (Pi_value/2.0) then -- [Pi/2, Pi/2]
				bx := Pi_value - bx
			elseif bx < (-Pi_value/2.0) then
				bx := - Pi_value - bx
			end
			check improved_bounds: bx >= -Pi_value/2.0 and bx <= Pi_value/2.0 end

			from
				n := 1
				term_sign := -1
			variant Sine_max + 2 - n until n > Sine_max loop
				term_sign := term_sign * -1
				Result := Result + ( ((bx ^ n)/fact(n)) * term_sign )
				n := n + 2
			end
		end

	cosine(x : DOUBLE) : DOUBLE is
			-- Portable cosine.
			-- Time: same as sine.
		do
			Result := sine(x+Pi_value/2)
		end

feature { NONE }
	
	Sine_max : INTEGER is 9

	mod(x,y : DOUBLE) : DOUBLE is
			-- Modulus (floating point).
		local
			sys : P_SYSTEM
		do
			!!sys
			Result := x - (y * sys.floor(x/y))
		end

	fact(n : INTEGER) : DOUBLE is
			-- Calculate precomputed n!
			-- n bounded, but n! grows very quickly and fact(n) overflows 
			-- DOUBLE on most platforms for n = 35.
			-- Time: O(1)
		require
			positive: n > 0
			maximum: n < 30
		do
			inspect n
			when 1 then Result := 1
			when 2 then Result := 2
			when 3 then Result := 6
			when 4 then Result := 24
			when 5 then Result := 120
			when 6 then Result := 720
			when 7 then Result := 5040
			when 8 then Result := 40320
			when 9 then Result := 362880
			when 10 then Result := 3628800
			when 11 then Result := 39916800
			when 12 then Result := 479001600
			when 13 then Result := 6.22702e+009
			when 14 then Result := 8.71783e+010
			when 15 then Result := 1.30767e+012
			when 16 then Result := 2.09228e+013
			when 17 then Result := 3.55687e+014
			when 18 then Result := 6.40237e+015
			when 19 then Result := 1.21645e+017
			when 20 then Result := 2.4329e+018
			when 21 then Result := 5.10909e+019
			when 22 then Result := 1.124e+021
			when 23 then Result := 2.5852e+022
			when 24 then Result := 6.20448e+023
			when 25 then Result := 1.55112e+025
			when 26 then Result := 4.03291e+026
			when 27 then Result := 1.08889e+028
			when 28 then Result := 3.04888e+029
			when 29 then Result := 8.84176e+030
			-- else exception
			end
		end

end
