-- Pylon: a foundation library; version 0.81(beta); 1997-05-08

-- 
-- Copyright (c) 1997 AlterSoft, England, E.U.
-- 
-- Permission is hereby granted, without written agreement and without
-- licence or royalty fees, to use, copy, modify, or distribute this source
-- code or binary programs using it or part of it, subject to the following
-- condition:
-- 
-- The above copyright notice and the following disclaimer must appear in
-- all copies or substantial portions of this source code. (Binary programs
-- derived from the source code can be distributed without restriction or
-- mention of the origin of the source.)
-- 
-- THIS SOFTWARE IS PROVIDED BY ALTERSOFT AND THE AUTHORS ``AS IS'' AND
-- ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED. IN NO EVENT SHALL ALTERSOFT OR THE AUTHORS BE LIABLE
-- TO ANY PARTY FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES ARISING IN ANY WAY OUT OF THE USE OF THIS SOURCE
-- CODE.
-- 

indexing
	cluster: Pylon, Container
	description: "Concrete indexed list, implemented by an array"
	comment: "Implemented with an ARRAY whose size is doubled (and reduced) automatically"
	interface: client

class P_SEQUENCE_LIST[G]

inherit
	P_LIST[G]
		end
	P_SEQUENCE[G]
		end

creation
	make

feature { NONE } -- Attributes

	container : ARRAY[G]
			-- actual container
	current_item : INTEGER
			-- current item

feature { NONE }
	
	Grow_factor : INTEGER is 2
	Default_size : INTEGER is 20

feature { NONE } -- Attribute for locking
	
	set_locked : P_HASH_SET[INTEGER]

feature -- Creation

	make, reset is
			-- Remove all items.
			-- Time: O(1)
		do
			!!container.make(1, Default_size)
			current_item := 0
			count := 0
			set_locked := Void
			make_traversable
		end
	
feature -- Iteration

	iterator : P_ITERATOR[G] is
			-- Get a two-way iterator on this container.
		local
			actual : P_SEQUENCE_ITERATOR[G]
		do
			!!actual.make(Current)
			Result := actual
		end

feature -- Access

	index : INTEGER is
			-- Current index.
			-- Time: O(1)
		do
			Result := current_item
		end

	set_index(i:INTEGER) is
			-- Set current item to position 'i'.
			-- Time: O(1)
		do
			current_item := i
		end
			
feature -- Status
	
	count : INTEGER;
			-- Quantity of items in the list.

	item : G is
			-- Get current item.
			-- Time: O(1)
		do
			check 
				valid_count: count > 0
				valid_item: current_item /= 0 
			end
			Result := container.item(current_item)
		end


feature -- Update
	
	add(citem : G) is
			-- Add item at the end of the list.
			-- Added item becomes current.
			-- Time: O(1) amortised
		do
			count := count + 1
			if count > container.upper then -- increase the size
				container.resize(1, container.upper * Grow_factor)
			end
			container.put(citem,count)
			current_item := count
		ensure then
			in_container: container.item(count) = citem
			count_ok: count = old count + 1
		end

	replace(citem : G) is
			-- Replace current item.
			-- Time: O(1)
		do
			container.put(citem,current_item)
		end

	insert(citem : G) is
			-- Insert item at current position, and push all the following items. 
			-- Time: O(count)
		local
			i : INTEGER
		do
			check not_locked: not has_locked end

			-- grow
			count := count + 1
			if count > container.upper then -- increase the size
				container.resize(1, container.upper * Grow_factor)
			end
			
			check no_overlap: count <= container.upper end
			
			-- push [current_item,count] +1
			from i := count variant i - current_item until i = current_item loop
				check valid_idx: i > current_item end
				container.put(container.item(i-1),i)
				i := i - 1 -- forth
			end

			-- replace current_item
			container.put(citem,current_item)
		ensure then
			done: item = citem
			count_ok: count = old count + 1
		end

	remove is
			-- Remove current item.
			-- Time: O(count) amortised
		local
			i : INTEGER
		do
			-- copy following items
			from
				i := current_item 
			variant count - i until i >= count loop
				container.put(container.item(i+1),i)
				i := i + 1
			end

			count := count - 1
			current_item := count

			-- resize down.
			if (count + Default_size) < (container.count // Grow_factor) then
				-- leave a Default_size margin to avoid extra resizes
				-- on a cycle of add/remove near the margin.
				container.resize(1, count + Default_size)
			end
		ensure then
			count_ok: count = old count - 1
		end

feature -- Lock status

	is_locked : BOOLEAN is
			-- Is current item locked?
			-- Time: O(1) amortised [dictionary search]
		do
			if set_locked /= Void then
				set_locked.search(current_item)
				Result := set_locked.found
			end
		end

feature { P_MUTATOR } -- Lock update

	lock(idx : INTEGER) is
			-- Lock current item.
			-- Time: O(1)
		do
			if set_locked = Void then
				!!set_locked.make
			end
			set_locked.add(idx)
		end

	unlock(idx : INTEGER) is
			-- Unlock current item.
			-- Time: O(1) amortised
		do
			set_locked.search(idx)
			check is_locked: set_locked.found = true end
			set_locked.remove
		end

feature -- Sorting
	
	sort is
			-- Selection sort.
			-- Time: O(target.count^2) comparisons; O(target.count) swaps
		local
			i,j,min : INTEGER;
			temp : G
		do
			if count > 1 then
				from
					i := 1
				invariant
					external_loop: i > 0 and i <= count
				variant
					count - i
				until
					i >= count
				loop
					from
						min := i
						j := i + 1
					invariant
						inner_loop: j > i and j <= count + 1
					variant
						count - j + 1
					until
						j > count
					loop
						if order_relation.ordered(container.item(j),container.item(min)) then
							min := j
						end

						-- next
						j := j + 1
					end

					-- swap
					temp := container.item(min)
					container.put(container.item(i),min)
					container.put(temp,i)

					-- next
					i := i + 1
				end
			end
		end

invariant
	valid_current: (not is_empty) implies (current_item >= 1 and current_item <= count)
	good_container: container.lower = 1
	big_container: count /= 0 implies container.upper >= count
	small_container: count /= 0 implies container.count // Grow_factor <= Default_size + count
	
end
