-- Pylon: a foundation library; version 0.81(beta); 1997-05-08

-- 
-- Copyright (c) 1997 AlterSoft, England, E.U.
-- 
-- Permission is hereby granted, without written agreement and without
-- licence or royalty fees, to use, copy, modify, or distribute this source
-- code or binary programs using it or part of it, subject to the following
-- condition:
-- 
-- The above copyright notice and the following disclaimer must appear in
-- all copies or substantial portions of this source code. (Binary programs
-- derived from the source code can be distributed without restriction or
-- mention of the origin of the source.)
-- 
-- THIS SOFTWARE IS PROVIDED BY ALTERSOFT AND THE AUTHORS ``AS IS'' AND
-- ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED. IN NO EVENT SHALL ALTERSOFT OR THE AUTHORS BE LIABLE
-- TO ANY PARTY FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-- CONSEQUENTIAL DAMAGES ARISING IN ANY WAY OUT OF THE USE OF THIS SOURCE
-- CODE.
-- 

indexing
	cluster: Pylon, Container
	description: "Abstract tree"
	interface: abstract

deferred class P_TREE[G]

inherit
	P_CONTAINER[G]
		end
	P_ONEWAY_TRAVERSABLE[G]
		rename
			iterator as linear_iterator -- linear_iterator is the default linear iterator
		end
		
feature -- Linear iteration

	preorder_iterator : P_ONEWAY_ITERATOR[G] is
			-- Linear preorder iterator.
		deferred
		ensure
			valid: Result /= Void
			is_shallow_copy: -- new iterator provided for each call
		end

	postorder_iterator : P_ONEWAY_ITERATOR[G] is
			-- Linear postorder iterator.
		deferred
		ensure
			valid: Result /= Void
			is_shallow_copy: -- new iterator provided for each call
		end

--	inorder_iterator : P_ITERATOR[G] is
--		BTREE ONLY
--		deferred
--		ensure
--			valid: Result /= Void
--			is_shallow_copy: -- new iterator provided for each call
--		end

	levelorder_iterator : P_ONEWAY_ITERATOR[G] is
			-- Linear level order iterator.
		deferred
		ensure
			valid: Result /= Void
			is_shallow_copy: -- new iterator provided for each call
		end
		
feature -- Modification

	add_left(new : G) is
			-- Add an item to the left of current item.
		require
			writable: is_writable
			has_item: not is_empty
			not_root: not is_root
		deferred
		ensure
			count: count = old count + 1
			keep_reference: item = new
		end

	add_right(new : G) is
			-- Add an item to the right of current item.
		require
			writable: is_writable
			has_item: not is_empty
			not_root: not is_root
		deferred
		ensure
			count: count = old count + 1
			keep_reference: item = new
		end

	add_root(new_root : G) is
			-- Add the root item.
		require
			writable: is_writable
			empty: is_empty
		deferred
		ensure
			count: count = 1
			keep_reference: item = new_root
			root: is_root
			sole_root: is_top and is_bottom and is_left and is_right
		end
	
	add_child(new : G) is
			-- Add a first child to the current item.
			-- Other children may be added with add_right/add_left
		require
			writable: is_writable
			has_item: not is_empty
			down: is_bottom
		deferred
		ensure
			count: count = old count + 1
			keep_reference: item = new
			unique_child: is_left and is_right
			still_down: is_leaf
		end

	replace(x : G) is
			-- Replace current item.
		require
			writable: is_writable
			has_item: not is_empty
		deferred
		ensure
			keep_reference: x = item
		end

	remove is
			-- Remove current item and its children.
			-- The tree shall not have locked items.
		require
			writable: is_writable
			has_item: not is_empty
			unlocked: not has_locked
		deferred
		ensure
			count: count = old count - 1
			moved: is_root
		end

feature -- Cursor
	
	is_left : BOOLEAN is
			-- Is the current item the first one?
		require
			meaningful: not is_empty
		deferred
		end

	is_right : BOOLEAN is
			-- Is the current item the last one?
		require
			meaningful: not is_empty
		deferred
		end

	is_top : BOOLEAN is
			-- Is root generation?
		require
			meaningful: not is_empty
		deferred
		end

	is_bottom : BOOLEAN is
			-- Has the current item not got any children?
		require
			meaningful: not is_empty
		deferred
		end

	left is
			-- Go backwards one slot.
		require
			meaningful: not is_empty
			not_left: not is_left
		deferred
		end

	right is
			-- Go forward one item.
		require
			meaningful: not is_empty
			not_right: not is_right
		deferred
		end

	up is
			-- Go one generation upwards.
		require
			meaningful: not is_empty
			not_up: not is_top
		deferred
		end

	down is
			-- Go one generation downwards.
		require
			meaningful: not is_empty
			not_down: not is_bottom
		deferred
		ensure
			position: is_left
		end

	leftmost is
			-- Go to the leftmost sibling.
		require
			meaningful: not is_empty
		deferred
		ensure
			done: is_left
		end

	rightmost is
			-- Go to the rightmost sibling.
		require
			meaningful: not is_empty
		deferred
		ensure
			done: is_right
		end

	root is
			-- Go to root node.
		require
			meaningful: not is_empty
		deferred
		ensure
			done: is_root
		end

feature -- Queries
	
	depth : INTEGER is
			-- Number of level from the root. Root depth is 0.
		require
			has_item: not is_empty
		deferred
		ensure
			positive: Result >= 0
			root_zero: is_root implies Result = 0
		end

	is_root : BOOLEAN is
			-- Is the current item the root?
		require
			has_item: not is_empty
		do
			Result := is_top and is_left and is_right
		ensure
			done: Result = (is_top and is_left and is_right)
		end

	is_leaf : BOOLEAN is
			-- Is the current item a leaf (node that has no child)?
		require
			has_item: not is_empty
		do
			Result := is_bottom
		ensure
			done: Result = is_bottom
		end

feature -- Conversions.

	subtree : like Current is
			-- Tree which has current node as root node.
		require
			not_empty: not is_empty
		deferred
		ensure
			done: Result /= Void and not Result.is_empty
			is_shallow_copy: -- New tree, old items.
		end
	
	siblings_list : P_LIST[G] is
			-- List of siblings of current item, including current item.
		require
			not_empty: not is_empty
		deferred
		ensure
			done: Result /= Void and then not Result.is_empty
			is_shallow_copy: -- List is a copy but contains real items.
		end

end
